#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include <sys/types.h>
#include <sys/socket.h>

#include <linux/ax25.h>

#include "axutils.h"
#include "axconfig.h"
#include "version.h"

int main(int argc, char *argv[])
{
	struct full_sockaddr_ax25 dest;
	struct full_sockaddr_ax25 src;
	int s, n, dlen, len, interval = 30;
	char *addr, *port, *message;
	
	while ((n = getopt(argc, argv, "t:v")) != -1) {
		switch (n) {
			case 't':
				interval = atoi(optarg);
				if (interval < 1) {
					fprintf(stderr, "beacon: interval must be greater than on minute\n");
					return 1;
				}
				break;
			case 'v':
				printf("beacon: %s\n", version);
				return 0;
			case '?':
			case ':':
				fprintf(stderr, "usage: beacon [-t interval] [-v] <port> <message>\n");
				return 1;
		}
	}

	if (optind == argc || optind == argc - 1) {
		fprintf(stderr, "usage: beacon [-t interval] [-v] <port> <message>\n");
		return 1;
	}

	port    = argv[optind];
	message = argv[optind + 1];
	
	if (ax25_config_load_ports() == 0) {
		fprintf(stderr, "beacon: no AX.25 ports defined\n");
		return 1;
	}

	if ((addr = ax25_config_get_addr(port)) == NULL) {
		fprintf(stderr, "beacon: invalid AX.25 port setting - %s\n", port);
		return 1;
	}

	dlen = convert_call("IDENT", &dest);
		
	if ((len = convert_call(addr, &src)) == -1) {
		fprintf(stderr, "beacon: unable to convert callsign '%s'\n", addr);
		return 1;
	}

	for (;;) {
		if ((s = socket(AF_AX25, SOCK_DGRAM, 0)) == -1) {
			perror("beacon: socket");
			return 1;
		}

		if (bind(s, (struct sockaddr *)&src, len) == -1) {
			perror("beacon: bind");
			return 1;
		}
		
		if (sendto(s, message, strlen(message), 0, (struct sockaddr *)&dest, dlen) == -1) {
			perror("beacon: sendto");
			return 1;
		}

		close(s);

		sleep(interval * 60);
	}

	return 0;
}
